import json
import logging
import typing
from dataclasses import dataclass

import pytest

import attr

from apitist.hooks import (
    PreparedRequestHook,
    PrepRequestDebugLoggingHook,
    PrepRequestInfoLoggingHook,
    RequestAttrsConverterHook,
    RequestDataclassConverterHook,
    RequestDebugLoggingHook,
    RequestHook,
    RequestInfoLoggingHook,
    ResponseAttrsConverterHook,
    ResponseDataclassConverterHook,
    ResponseDebugLoggingHook,
    ResponseHook,
    ResponseInfoLoggingHook,
)
from apitist.requests import Session, session


@attr.s
class ExampleData:
    test: str = attr.ib()


@attr.s
class ExampleResponse:
    args: typing.Dict = attr.ib()
    data: str = attr.ib()
    files: typing.Dict = attr.ib()
    form: typing.Dict = attr.ib()
    headers: typing.Dict = attr.ib()
    json: typing.Any = attr.ib()
    origin: str = attr.ib()
    url: str = attr.ib()


@dataclass
class ExampleDataclass:
    test: str


@dataclass
class ExampleResponseDataclass:
    args: typing.Dict
    data: str
    files: typing.Dict
    form: typing.Dict
    headers: typing.Dict
    json: typing.Any
    origin: str
    url: str


class TestHooks:
    def test_session_creation(self):
        assert isinstance(session(), Session)

    @pytest.mark.parametrize(
        "hook", [(PreparedRequestHook), (RequestHook), (ResponseHook)]
    )
    def test_empty_hooks(self, hook):
        assert hook().run(None) is None

    @pytest.mark.parametrize("hook", [int, str, bool])
    def test_incorrect_hook(self, session, hook):
        session.add_hook(hook)
        assert len(session.request_hooks) == 0
        assert len(session.prep_request_hooks) == 0
        assert len(session.response_hooks) == 0

    @pytest.mark.usefixtures("enable_debug_logging")
    @pytest.mark.parametrize(
        "hook,text",
        [
            (RequestDebugLoggingHook, "Request"),
            (PrepRequestDebugLoggingHook, "Request"),
            (ResponseDebugLoggingHook, "Response"),
        ],
    )
    def test_debug_logging_hooks(self, session, hook, text, capture):
        session.add_hook(hook)
        session.get("http://httpbin.org/get")
        assert len(capture.records) == 2
        assert capture.records[1].levelno == logging.DEBUG
        assert text in capture.records[1].msg

    @pytest.mark.parametrize(
        "hook,text",
        [
            (RequestInfoLoggingHook, "Request"),
            (PrepRequestInfoLoggingHook, "Request"),
            (ResponseInfoLoggingHook, "Response"),
        ],
    )
    def test_info_logging_hooks(self, session, hook, text, capture):
        session.add_hook(hook)
        session.get("http://httpbin.org/get")
        assert len(capture.records) == 1
        assert capture.records[0].levelno == logging.INFO
        assert text in capture.records[0].msg

    def test_request_converter_attrs_class(self, session):
        session.add_hook(RequestAttrsConverterHook)
        res = session.get("http://httpbin.org/get", data=ExampleData("test"))
        assert res.request.body == json.dumps({"test": "test"}).encode("utf-8")

    def test_request_converter_non_attrs(self, session):
        session.add_hook(RequestAttrsConverterHook)
        res = session.get("http://httpbin.org/get", data="test")
        assert res.request.body == "test"

    def test_response_converter_adding_function(self, session):
        session.add_hook(ResponseAttrsConverterHook)
        res = session.get("http://httpbin.org/get")
        assert getattr(res, "structure")

    def test_response_converter_correct_type(self, session):
        session.add_hook(ResponseAttrsConverterHook)
        res = session.post("http://httpbin.org/post")
        res.structure(ExampleResponse)
        assert isinstance(res.data, ExampleResponse)

    def test_response_converter_incorrect_type(self, session):
        session.add_hook(ResponseAttrsConverterHook)
        res = session.post("http://httpbin.org/post")
        with pytest.raises(TypeError):
            res.structure(ExampleData)

    def test_request_converter_dataclass_class(self, session):
        session.add_hook(RequestDataclassConverterHook)
        res = session.get(
            "http://httpbin.org/get", data=ExampleDataclass("test")
        )
        assert res.request.body == json.dumps({"test": "test"}).encode("utf-8")

    def test_request_converter_non_dataclass(self, session):
        session.add_hook(RequestDataclassConverterHook)
        res = session.get("http://httpbin.org/get", data="test")
        assert res.request.body == "test"

    def test_response_converter_adding_function_dataclass(self, session):
        session.add_hook(ResponseDataclassConverterHook)
        res = session.get("http://httpbin.org/get").vr()
        assert getattr(res, "structure")
        assert getattr(res, "verify_response")
        assert getattr(res, "vr")

    def test_response_verify_response_success(self, session):
        session.get("http://httpbin.org/get").vr([200, 201])

    def test_response_verify_response_error(self, session):
        with pytest.raises(ValueError):
            session.get("http://httpbin.org/get").vr(400)

    def test_response_converter_correct_dataclass_type(self, session):
        session.add_hook(ResponseDataclassConverterHook)
        res = session.post("http://httpbin.org/post")
        res.structure(ExampleResponseDataclass)
        assert isinstance(res.data, ExampleResponseDataclass)

    def test_response_converter_incorrect_dataclass_type(self, session):
        session.add_hook(ResponseDataclassConverterHook)
        res = session.post("http://httpbin.org/post")
        with pytest.raises(TypeError):
            res.structure(ExampleDataclass)



import logging

import pytest

from testfixtures import LogCapture

from apitist.constructor import Converter, ConverterType
from apitist.logging import Logging
from apitist.random import Randomer
from apitist.requests import Session


@pytest.fixture(params=[ConverterType.ATTRS, ConverterType.DATACLASS])
def converter(request):
    return Converter(request.param)


@pytest.fixture
def session():
    return Session()


@pytest.fixture
def randomer():
    return Randomer()


@pytest.fixture
def random():
    rand = Randomer()
    rand.add_type(str, lambda: "test")
    rand.add_type(int, lambda: 50)
    return rand


@pytest.fixture()
def capture():
    with LogCapture() as capture:
        yield capture


@pytest.fixture()
def enable_debug_logging():
    Logging.logger.setLevel(logging.DEBUG)
    yield
    Logging.logger.setLevel(Logging.LOG_LEVEL)



import inspect
from abc import ABC
from typing import List, Type, TypeVar, Union

from requests import PreparedRequest, Request, Response
from requests import Session as OldSession

from apitist.logging import Logging


class SessionHook(ABC):
    pass


class RequestHook(SessionHook):
    def run(self, request: Request) -> Request:
        ...


class PreparedRequestHook(SessionHook):
    def run(self, request: PreparedRequest) -> PreparedRequest:
        ...


class ResponseHook(SessionHook):
    def run(self, response: Response) -> Response:
        ...


DataType = TypeVar("DataType")


class ApitistResponse(Response):
    data: DataType

    def __init__(self, response: Response):
        self.__dict__ = response.__dict__

    def verify_response(
        self, ok_status: Union[int, List[int]] = 200
    ) -> "ApitistResponse":
        func = inspect.stack()[2][3]
        if isinstance(ok_status, int):
            ok_status = [ok_status]
        if self.status_code not in ok_status:
            raise ValueError(
                f"Verified response: function {func} failed: "
                f"server responded {self.status_code} "
                f"with data: {self.content}"
            )
        else:
            Logging.logger.info(
                f"Verified response: function {func} code {self.status_code}"
            )
        return self

    def vr(self, ok_status: Union[int, List[int]] = 200) -> "ApitistResponse":
        return self.verify_response(ok_status=ok_status)

    def structure(self, t: Type[DataType]) -> "ApitistResponse":
        ...


class Session(OldSession):
    def __init__(self):
        super().__init__()
        self.request_hooks = []
        self.prep_request_hooks = []
        self.response_hooks = []

    def _add_hook(
        self,
        hook: Type[Union[RequestHook, PreparedRequestHook, ResponseHook]],
        array: list,
    ):
        array.append(hook)

    def add_request_hook(self, hook: Type[RequestHook]):
        Logging.logger.debug("Adding new request hook")
        self._add_hook(hook, self.request_hooks)

    def add_prep_request_hook(self, hook: Type[PreparedRequestHook]):
        Logging.logger.debug("Adding new prepared request hook")
        self._add_hook(hook, self.prep_request_hooks)

    def add_response_hook(self, hook: Type[ResponseHook]):
        Logging.logger.debug("Adding new response hook")
        self._add_hook(hook, self.response_hooks)

    def add_hook(
        self, hook: Type[Union[RequestHook, PreparedRequestHook, ResponseHook]]
    ):
        if issubclass(hook, RequestHook):
            self.add_request_hook(hook)
        elif issubclass(hook, PreparedRequestHook):
            self.add_prep_request_hook(hook)
        elif issubclass(hook, ResponseHook):
            self.add_response_hook(hook)

    def _run_hooks(
        self,
        array: List[
            Type[Union[RequestHook, PreparedRequestHook, ResponseHook]]
        ],
        data: Union[Request, PreparedRequest, Response, ApitistResponse],
    ):
        for hook in array:
            data = hook().run(data)
        return data

    def request(
        self,
        method,
        url,
        params=None,
        data=None,
        headers=None,
        cookies=None,
        files=None,
        auth=None,
        timeout=None,
        allow_redirects=True,
        proxies=None,
        hooks=None,
        stream=None,
        verify=None,
        cert=None,
        json=None,
    ) -> ApitistResponse:
        """Constructs a :class:`Request <Request>`, prepares it and sends it.
        Returns :class:`Response <Response>` object.

        :param method: method for the new :class:`Request` object.
        :param url: URL for the new :class:`Request` object.
        :param params: (optional) Dictionary or bytes to be sent in the query
            string for the :class:`Request`.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the
            :class:`Request`.
        :param headers: (optional) Dictionary of HTTP Headers to send with the
            :class:`Request`.
        :param cookies: (optional) Dict or CookieJar object to send with the
            :class:`Request`.
        :param files: (optional) Dictionary of ``'filename':
            file-like-objects`` for multipart encoding upload.
        :param auth: (optional) Auth tuple or callable to enable
            Basic/Digest/Custom HTTP Auth.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :type timeout: float or tuple
        :param allow_redirects: (optional) Set to True by default.
        :type allow_redirects: bool
        :param proxies: (optional) Dictionary mapping protocol or protocol and
            hostname to the URL of the proxy.
        :param stream: (optional) whether to immediately download the response
            content. Defaults to ``False``.
        :param verify: (optional) Either a boolean, in which case it controls
            whether we verify the server's TLS certificate, or a string,
            in which case it must be a path to a CA bundle to use.
            Defaults to ``True``.
        :param cert: (optional) if String, path to ssl client cert file (.pem).
            If Tuple, ('cert', 'key') pair.
        :rtype: requests.Response
        """
        # Create the Request.
        req = Request(
            method=method.upper(),
            url=url,
            headers=headers,
            files=files,
            data=data or {},
            json=json,
            params=params or {},
            auth=auth,
            cookies=cookies,
            hooks=hooks,
        )
        req = self._run_hooks(self.request_hooks, req)
        prep = self.prepare_request(req)
        prep = self._run_hooks(self.prep_request_hooks, prep)

        proxies = proxies or {}

        settings = self.merge_environment_settings(
            prep.url, proxies, stream, verify, cert
        )

        # Send the request.
        send_kwargs = {"timeout": timeout, "allow_redirects": allow_redirects}
        send_kwargs.update(settings)
        resp = ApitistResponse(self.send(prep, **send_kwargs))
        resp = self._run_hooks(self.response_hooks, resp)

        return resp


def session():
    """
    Returns a :class:`Session` for context-management.

    :rtype: Session
    """
    return Session()



from dataclasses import MISSING, dataclass, field

import pytest

import attr
import pendulum

from apitist.constructor import (
    ConverterType,
    MissingDict,
    NothingDict,
    _structure_date_time,
    _subclass,
    _unstructure_date_time,
)


class TypeStr(str):
    pass


@attr.s
class ExampleData:
    test: TypeStr = attr.ib()
    example: pendulum.DateTime = attr.ib(default=None)


@dataclass
class ExampleDataclass:
    test: TypeStr
    example: pendulum.DateTime = field(default=None)


class TestConstructor:
    @pytest.mark.parametrize(
        "t,verify_type,result",
        [
            (str, str, True),
            (str, TypeStr, True),
            (int, str, False),
            (int, TypeStr, False),
        ],
    )
    def test_subclass(self, t, verify_type, result):
        assert _subclass(t)(verify_type) == result

    def test_structure_date_time(self):
        isostring = "2019-03-06T14:58:10.123+0500"
        datetime = _structure_date_time(isostring, None)
        assert datetime

    @pytest.mark.parametrize("value", [0, {}, [], None])
    def test_structure_date_time_none(self, value):
        datetime = _structure_date_time(value, None)
        assert datetime is None

    @pytest.mark.parametrize(
        "datetime,result",
        [
            (
                pendulum.parse("2019-03-06T14:58:10.123+0500"),
                "2019-03-06T14:58:10.123000+05:00",
            ),
            (
                pendulum.parse("2019-03-06T14:58:10"),
                "2019-03-06T14:58:10+00:00",
            ),
            (pendulum.parse("2019-03-06"), "2019-03-06T00:00:00+00:00"),
        ],
    )
    def test_unstructure_date_time(self, datetime, result):
        datetime = _unstructure_date_time(datetime)
        assert datetime == result

    @pytest.mark.parametrize("value", [0, {}, [], None])
    def test_unstructure_date_time_none(self, value):
        datetime = _unstructure_date_time(value)
        assert datetime is None

    def test_converter_additional_hooks(self, converter):
        data = {"test": TypeStr("test"), "example": "2019-03-06T14:58:10"}
        _type = (
            ExampleData
            if converter._converter_type == ConverterType.ATTRS
            else ExampleDataclass
        )
        with pytest.raises(ValueError):
            converter.structure(data, _type)
        converter.register_additional_hooks()
        assert converter.structure(data, _type)

    @pytest.mark.parametrize(
        "value,result", [("test", True), (0, True), (attr.NOTHING, False)]
    )
    def test_nothing_dict_set_item(self, value, result):
        dict_ = NothingDict()
        dict_["test"] = value
        assert bool(dict_) == result

    @pytest.mark.parametrize(
        "value,result", [("test", True), (0, True), (MISSING, False)]
    )
    def test_missing_dict_set_item(self, value, result):
        dict_ = MissingDict()
        dict_["test"] = value
        assert bool(dict_) == result

    @pytest.mark.parametrize(
        "value,result", [("test", "test"), (0, "0"), (None, None)]
    )
    def test_structure_call(self, converter, value, result):
        assert converter._structure_call(value, str) == result



import sys
from typing import (
    Dict,
    FrozenSet,
    Mapping,
    MutableSequence,
    MutableSet,
    Sequence,
    Tuple,
)

version_info = sys.version_info[0:3]
is_py37_or_higher = version_info[:2] in [(3, 7), (3, 8)]


def is_attrs_class(cls):
    return getattr(cls, "__attrs_attrs__", None) is not None


def has_args(cls):
    return getattr(cls, "__args__", None) is not None


def _subclass(typ):
    """ a shortcut """
    return lambda cls: issubclass(cls, typ)


if is_py37_or_higher:
    from typing import List, Union, _GenericAlias

    def is_union_type(obj):
        return (
            obj is Union
            or isinstance(obj, _GenericAlias)
            and obj.__origin__ is Union
        )

    def is_sequence(type):
        return (
            type is List
            or type.__class__ is _GenericAlias
            and issubclass(type.__origin__, Sequence)
        )

    def is_frozenset(type):
        return type.__class__ is _GenericAlias and issubclass(
            type.__origin__, FrozenSet
        )

    def is_mutable_set(type):
        return type.__class__ is _GenericAlias and issubclass(
            type.__origin__, MutableSet
        )

    def is_tuple(type):
        return type is Tuple or (
            type.__class__ is _GenericAlias
            and issubclass(type.__origin__, Tuple)
        )

    def is_mapping(type):
        return type is Mapping or (
            type.__class__ is _GenericAlias
            and issubclass(type.__origin__, Mapping)
        )

    bare_list_args = List.__args__
    bare_seq_args = Sequence.__args__
    bare_mapping_args = Mapping.__args__
    bare_dict_args = Dict.__args__
    bare_mutable_seq_args = MutableSequence.__args__

    def is_bare(type):
        args = type.__args__
        return (
            args == bare_list_args
            or args == bare_seq_args
            or args == bare_mapping_args
            or args == bare_dict_args
            or args == bare_mutable_seq_args
        )


else:
    from typing import _Union

    def is_union_type(obj):
        """Return true if the object is a union. """
        return isinstance(obj, _Union)

    def is_frozenset(type):
        return issubclass(type, FrozenSet)

    def is_mutable_set(type):
        return issubclass(type, MutableSet)

    def is_sequence(type):
        return issubclass(type, Sequence)

    def is_tuple(type):
        return issubclass(type, Tuple)

    def is_mapping(type):
        return issubclass(type, Mapping)



import logging
import typing
from dataclasses import dataclass, field, is_dataclass

import pytest

import attr

from apitist.constructor import convclass, converter


class TypeStr(str):
    pass


@attr.s
class Child:
    value1: int = attr.ib()
    value2: typing.List[str] = attr.ib()
    value3: typing.Tuple[str] = attr.ib(default=tuple())
    value4: str = attr.ib(default=None)
    value5: typing.Union[int] = attr.ib(default=None)


@dataclass
class ChildDataclass:
    value1: int
    value2: typing.List[str]
    value3: typing.Tuple[str] = field(default_factory=tuple)
    value4: str = field(default=None)
    value5: typing.Union[int] = field(default=None)


@attr.s
class Data:
    val1: str = attr.ib()
    val2: Child = attr.ib()
    val3: str = attr.ib(default=None)


@dataclass
class DataDataclass:
    val1: str
    val2: ChildDataclass
    val3: str = field(default=None)


class TestRandomer:
    def test_add_type(self, randomer):
        def func():
            return "test"

        randomer.add_type(str, func)
        assert randomer.available_hooks == {str: func}

    def test_add_type_exists(self, randomer, capture):
        def func():
            return "test"

        randomer.add_type(str, func)
        randomer.add_type(str, func)
        assert capture.records[0].levelno == logging.WARN

    def test_add_types(self, randomer):
        def func():
            return "test"

        randomer.add_types({str: func})
        assert randomer.available_hooks == {str: func}

    def test_get_hook_exists(self, random):
        assert random.get_hook(str)

    def test_get_hook_not_exists(self, random):
        with pytest.raises(TypeError):
            random.get_hook(list)

    def test_run_hook_exists(self, random):
        assert random.run_hook(str)

    def test_run_hook_not_exists(self, random):
        with pytest.raises(TypeError):
            random.run_hook(list)

    def test_run_hook_one_subclass(self, randomer):
        def func():
            return "test"

        def func2():
            return "test2"

        randomer.add_type(str, func)
        randomer.add_type(TypeStr, func2)
        assert randomer.run_hook(str) == "test"
        assert randomer.run_hook(TypeStr) == "test2"

    @pytest.mark.parametrize(
        "key,value",
        [
            ("str", "str"),
            ("int", 10),
            ("list", ["some", "data"]),
            ("dict", {"key": 100, "val": {10, 12, 13}}),
        ],
    )
    def test_run_hook_kwargs_pass(self, randomer, key, value):
        def func(**kwargs):
            if kwargs.get(key):
                return kwargs.get(key)
            return "test"

        randomer.add_type(str, func)
        assert randomer.run_hook(str) == "test"
        assert randomer.run_hook(str, **{key: value}) == value

    @pytest.mark.parametrize(
        "type_,object_",
        [
            (Child, Child(50, ["test"], ("test",), "test", 50)),
            (
                ChildDataclass,
                ChildDataclass(50, ["test"], ("test",), "test", 50),
            ),
            (
                Data,
                Data(
                    "test", Child(50, ["test"], ("test",), "test", 50), "test"
                ),
            ),
            (
                DataDataclass,
                DataDataclass(
                    "test",
                    ChildDataclass(50, ["test"], ("test",), "test", 50),
                    "test",
                ),
            ),
        ],
    )
    def test_random_object(self, random, type_, object_):
        assert random.object(type_) == object_

    def test_random_object_missing_type(self, random):
        assert random.object(dict) is None

    @pytest.mark.parametrize(
        "type_,object_",
        [
            (Child, Child(50, ["test"], tuple(), None)),
            (ChildDataclass, ChildDataclass(50, ["test"], tuple(), None)),
            (Data, Data("test", Child(50, ["test"], tuple(), None), None)),
            (
                DataDataclass,
                DataDataclass(
                    "test", ChildDataclass(50, ["test"], tuple(), None), None
                ),
            ),
        ],
    )
    def test_random_object_required_only(self, random, type_, object_):
        assert random.object(type_, required_only=True) == object_

    @pytest.mark.parametrize(
        "type_,ignore,object_",
        [
            (Child, ["value1"], Child(None, ["test"], ("test",), "test", 50)),
            (
                ChildDataclass,
                ["value1"],
                ChildDataclass(None, ["test"], ("test",), "test", 50),
            ),
            (
                Data,
                ["value1", "val1"],
                Data(
                    None, Child(None, ["test"], ("test",), "test", 50), "test"
                ),
            ),
            (
                DataDataclass,
                ["value1", "val1"],
                DataDataclass(
                    None,
                    ChildDataclass(None, ["test"], ("test",), "test", 50),
                    "test",
                ),
            ),
            (Data, ["value1", "val2"], Data("test", None, "test")),
            (
                DataDataclass,
                ["value1", "val2"],
                DataDataclass("test", None, "test"),
            ),
        ],
    )
    def test_random_object_ignore(self, random, type_, ignore, object_):
        assert random.object(type_, ignore=ignore) == object_

    @pytest.mark.parametrize(
        "type_,ignore,object_",
        [
            (Child, ["value1"], Child(50, None, tuple(), None)),
            (
                ChildDataclass,
                ["value1"],
                ChildDataclass(50, None, tuple(), None),
            ),
            (Data, ["value1", "val1"], Data("test", None, None)),
            (
                DataDataclass,
                ["value1", "val1"],
                DataDataclass("test", None, None),
            ),
            (
                Data,
                ["value1", "val2"],
                Data(None, Child(50, None, tuple(), None), None),
            ),
            (
                DataDataclass,
                ["value1", "val2"],
                DataDataclass(
                    None, ChildDataclass(50, None, tuple(), None), None
                ),
            ),
        ],
    )
    def test_random_object_ignore_inverse(
        self, random, type_, ignore, object_
    ):
        assert random.object(type_, ignore=ignore, inverse=True) == object_

    @pytest.mark.parametrize(
        "type_,ignore,object_",
        [
            (Child, ["value1"], Child(50, None, tuple(), None)),
            (
                ChildDataclass,
                ["value1"],
                ChildDataclass(50, None, tuple(), None),
            ),
            (Data, ["value1", "val1"], Data("test", None, None)),
            (
                DataDataclass,
                ["value1", "val1"],
                DataDataclass("test", None, None),
            ),
            (
                Data,
                ["value1", "val2"],
                Data(None, Child(50, None, tuple(), None), None),
            ),
            (
                DataDataclass,
                ["value1", "val2"],
                DataDataclass(
                    None, ChildDataclass(50, None, tuple(), None), None
                ),
            ),
        ],
    )
    def test_random_object_only(self, random, type_, ignore, object_):
        assert random.object(type_, only=ignore) == object_

    def test_random_object_only_ignore(self, random):
        with pytest.raises(ValueError):
            random.object(Data, only=["value1"], ignore=["value2"])

    @pytest.mark.parametrize(
        "type_,object_,params",
        [
            (
                Child,
                Child(560, ["test"], ("test",), "test", 50),
                {"value1": 560},
            ),
            (
                ChildDataclass,
                ChildDataclass(560, ["test"], ("test",), "test", 50),
                {"value1": 560},
            ),
            (
                Data,
                Data(
                    "rest", Child(560, ["test"], ("test",), "test", 50), "test"
                ),
                {"value1": 560, "val1": "rest"},
            ),
            (
                DataDataclass,
                DataDataclass(
                    "rest",
                    ChildDataclass(560, ["test"], ("test",), "test", 50),
                    "test",
                ),
                {"value1": 560, "val1": "rest"},
            ),
        ],
    )
    def test_random_object_set_params(self, random, type_, object_, params):
        assert random.object(type_, **params) == object_

    def test_random_object_union(self, random):
        @attr.s
        class NewObj:
            value1: typing.Union[int, str] = attr.ib()
            value2: typing.Optional[str] = attr.ib()

        @dataclass
        class NewObjDataclass:
            value1: typing.Union[int, str]
            value2: typing.Optional[str]

        for _ in range(20):
            assert converter.unstructure(random.object(NewObj)).get(
                "value1"
            ) in ["test", 50]
            assert converter.unstructure(random.object(NewObj)).get(
                "value2"
            ) in ["test", None]
            assert convclass.unstructure(random.object(NewObjDataclass)).get(
                "value1"
            ) in ["test", 50]
            assert convclass.unstructure(random.object(NewObjDataclass)).get(
                "value2"
            ) in ["test", None]

    @pytest.mark.parametrize(
        "type_,use,exp_data",
        [
            (Child, ["value1"], {"value1": 50}),
            (Child, ["value2"], {"value2": ["test"]}),
            (Child, ["value3"], {"value3": ("test",)}),
            (Child, ["value4"], {"value4": "test"}),
            (Child, ["value5"], {"value5": 50}),
            (ChildDataclass, ["value1"], {"value1": 50}),
            (ChildDataclass, ["value2"], {"value2": ["test"]}),
            (ChildDataclass, ["value3"], {"value3": ("test",)}),
            (ChildDataclass, ["value4"], {"value4": "test"}),
            (ChildDataclass, ["value5"], {"value5": 50}),
            (Data, ["val1"], {"val1": "test"}),
            (
                Data,
                ["val1", "value2"],
                {"val1": "test", "val2": {"value2": ["test"]}},
            ),
            (DataDataclass, ["val1"], {"val1": "test"}),
            (
                DataDataclass,
                ["val1", "value2"],
                {"val1": "test", "val2": {"value2": ["test"]}},
            ),
        ],
    )
    def test_random_partial(self, random, type_, use, exp_data):
        data = random.partial(type_, use=use)
        if is_dataclass(data):
            assert convclass.unstructure(data) == exp_data
        else:
            assert converter.unstructure(data) == exp_data

    def test_random_partial_missing_type(self, random):
        assert random.partial(dict) is None

    @pytest.mark.parametrize(
        "type_,use,params,exp_data",
        [
            (
                Child,
                ["value1"],
                {"value2": "rest"},
                {"value1": 50, "value2": "rest"},
            ),
            (
                ChildDataclass,
                ["value1"],
                {"value2": "rest"},
                {"value1": 50, "value2": "rest"},
            ),
            (
                Data,
                ["val1"],
                {"value2": "rest"},
                {"val1": "test", "val2": {"value2": "rest"}},
            ),
            (
                DataDataclass,
                ["val1"],
                {"value2": "rest"},
                {"val1": "test", "val2": {"value2": "rest"}},
            ),
        ],
    )
    def test_random_partial_set_param(
        self, random, type_, use, params, exp_data
    ):
        data = random.partial(type_, use=use, **params)
        if is_dataclass(data):
            assert convclass.unstructure(data) == exp_data
        else:
            assert converter.unstructure(data) == exp_data

    @pytest.mark.parametrize(
        "use,exp_data",
        [
            (["value4"], {"value4": "test"}),
            (
                ["value3", "value6"],
                {"value5": {"value3": {"data": 120}}, "value6": {"data": 120}},
            ),
        ],
    )
    def test_random_partial_attrs_custom_hook(self, random, use, exp_data):
        @attr.s
        class Child2:
            value1: int = attr.ib()

        @attr.s
        class Child1:
            value2: int = attr.ib()
            value3: Child2 = attr.ib()

        @attr.s
        class Parent:
            value4: str = attr.ib()
            value5: Child1 = attr.ib()
            value6: Child2 = attr.ib()

        random.add_type(Child2, lambda: {"data": 120})

        data = random.partial(Parent, use=use)
        assert converter.unstructure(data) == exp_data

    @pytest.mark.parametrize(
        "use,exp_data",
        [
            (["value4"], {"value4": "test"}),
            (
                ["value3", "value6"],
                {"value5": {"value3": {"data": 120}}, "value6": {"data": 120}},
            ),
        ],
    )
    def test_random_partial_dataclasses_custom_hook(
        self, random, use, exp_data
    ):
        @dataclass
        class Child2:
            value1: int

        @dataclass
        class Child1:
            value2: int
            value3: Child2

        @dataclass
        class Parent:
            value4: str
            value5: Child1
            value6: Child2

        random.add_type(Child2, lambda: {"data": 120})

        data = random.partial(Parent, use=use)
        assert convclass.unstructure(data) == exp_data

    def test_random_partial_union(self, random):
        @attr.s
        class NewObj:
            value1: typing.Union[int, str] = attr.ib()
            value2: int = attr.ib()

        @dataclass
        class NewObjDataclass:
            value1: typing.Union[int, str]
            value2: int

        for _ in range(10):
            assert converter.unstructure(
                random.partial(NewObj, use=["value1"])
            ) in [{"value1": 50}, {"value1": "test"}]
            assert convclass.unstructure(
                random.partial(NewObjDataclass, use=["value1"])
            ) in [{"value1": 50}, {"value1": "test"}]



from dataclasses import MISSING
from enum import Enum

import attr
import cattr
import convclasses
import pendulum

from apitist.utils import _subclass


def _structure_date_time(isostring, _):
    """Structure hook for :class:`pendulum.DateTime`"""
    if isinstance(isostring, str):
        return pendulum.parse(isostring)
    else:
        return None


def _unstructure_date_time(dt):
    """Unstructure hook for :class:`pendulum.DateTime`"""
    if isinstance(dt, pendulum.DateTime):
        return dt.to_rfc3339_string()
    else:
        return None


class ConverterType(Enum):
    ATTRS = "attrs"
    DATACLASS = "dataclass"


class _Converter:
    """Converts between structured and unstructured data."""

    def set_dict_factory(self, dict_factory):
        self._dict_factory = dict_factory

    def register_hooks(self, cls, structure, unstructure):
        """
        Register primitive-to-class and class-to-primitive converter
        functions for a class.

        The structure function should take two arguments:
          * a Python object to be converted,
          * the type to convert to

        and return the instance of the class. The type may seem redundant, but
        is sometimes needed (for example, when dealing with generic classes).

        The unstructure function should take an instance of the class and
        return its Python equivalent.

        """
        self.register_structure_hook(cls, structure)
        self.register_unstructure_hook(cls, unstructure)

    def register_hooks_funcs(self, check_func, structure, unstructure):
        """
        Register primitive-to-class and class-to-primitive converter functions
        for a class, using a function to check if it's a match.
        """
        self.register_structure_hook_func(check_func, structure)
        self.register_unstructure_hook_func(check_func, unstructure)

    def register_additional_hooks(self):
        """
        Register additional hooks:

        * :class:`str` - all its instances would be structured
          and unstructured as :class:`str`
        * :class:`pendulum.DateTime` - datetime would be parsed using pendulum
          and unstructured in RFC3339 format
        """
        self.register_hooks_funcs(
            _subclass(str), self._unstructure_identity, self._structure_call
        )
        self.register_hooks(
            pendulum.DateTime, _structure_date_time, _unstructure_date_time
        )

    def _structure_call(self, obj, cl):
        """Just call ``cl`` with the given ``obj``."""
        if obj is None:
            return None
        return cl(obj)


class AttrsConverter(_Converter, cattr.Converter):
    _converter_type = ConverterType.ATTRS


class DataclassConverter(_Converter, convclasses.Converter):
    _converter_type = ConverterType.DATACLASS


class NothingDict(dict):
    """
    Default dict for unstructuring

    It is used for unstructuring Type with ignoring some fields.
    If given field is :class:`attr.NOTHING` - it would not be unstructured in
    dict.
    """

    def __setitem__(self, key, value):
        if value == attr.NOTHING:
            return
        super().__setitem__(key, value)


class MissingDict(dict):
    """
    Default dict for unstructuring

    It is used for unstructuring Type with ignoring some fields.
    If given field is :class:`dataclasses.MISSING` -
    it would not be unstructured in dict.
    """

    def __setitem__(self, key, value):
        if value == MISSING:
            return
        super().__setitem__(key, value)


def Converter(converter_type: ConverterType = ConverterType.ATTRS):
    if converter_type == ConverterType.ATTRS:
        return AttrsConverter()
    else:
        return DataclassConverter()


converter = Converter()
converter.set_dict_factory(NothingDict)
converter.register_additional_hooks()

convclass = Converter(ConverterType.DATACLASS)
convclass.set_dict_factory(MissingDict)
convclass.register_additional_hooks()



import dataclasses
import types
from typing import Type

import attr
from requests import PreparedRequest, Request, Response

from apitist.utils import is_attrs_class

from .constructor import convclass, converter
from .logging import Logging
from .requests import PreparedRequestHook, RequestHook, ResponseHook


class RequestDebugLoggingHook(RequestHook):
    formatter = "Request {req.method} {req.url} {req.data}"

    def run(self, request: Request) -> Request:
        Logging.logger.debug(self.formatter.format(req=request))
        return request


class RequestInfoLoggingHook(RequestHook):
    formatter = "Request {req.method} {req.url} {req.data}"

    def run(self, request: Request) -> Request:
        Logging.logger.info(self.formatter.format(req=request))
        return request


class PrepRequestDebugLoggingHook(PreparedRequestHook):
    formatter = "Request {req.method} {req.url} {req.body}"

    def run(self, request: PreparedRequest) -> PreparedRequest:
        Logging.logger.debug(self.formatter.format(req=request))
        return request


class PrepRequestInfoLoggingHook(PreparedRequestHook):
    formatter = "Request {req.method} {req.url} {req.body}"

    def run(self, request: PreparedRequest) -> PreparedRequest:

        Logging.logger.info(self.formatter.format(req=request))
        return request


class ResponseDebugLoggingHook(ResponseHook):
    formatter = (
        "Response {res.status_code} {res.request.method} "
        "{res.url} {res.content}"
    )

    def run(self, response: Response) -> Response:
        Logging.logger.debug(self.formatter.format(res=response))
        return response


class ResponseInfoLoggingHook(ResponseHook):
    formatter = (
        "Response {res.status_code} {res.request.method} "
        "{res.url} {res.content}"
    )

    def run(self, response: Response) -> Response:
        Logging.logger.info(self.formatter.format(res=response))
        return response


class RequestAttrsConverterHook(RequestHook):
    def run(self, request: Request) -> Request:
        if is_attrs_class(request.data):
            request.json = converter.unstructure(request.data)
            request.data = None
        return request


class RequestDataclassConverterHook(RequestHook):
    def run(self, request: Request) -> Request:
        if dataclasses.is_dataclass(request.data):
            request.json = convclass.unstructure(request.data)
            request.data = None
        return request


RequestConverterHook = RequestAttrsConverterHook


class ResponseAttrsConverterHook(ResponseHook):
    def run(self, response: Response) -> Response:
        def func(self, t: Type) -> Response:
            try:
                self.data = converter.structure(self.json(), t)
            except TypeError as e:
                fields = attr.fields_dict(t)
                raise TypeError(
                    f"Got miss-matched parameters in dicts. "
                    f"Info about first level:"
                    f"\n\tExpect: {sorted(fields)}"
                    f"\n\tActual: {sorted(dict(self.json()).keys())}"
                    f"\n\nOriginal exception: {e}"
                )
            return self

        response.structure = types.MethodType(func, response)
        return response


class ResponseDataclassConverterHook(ResponseHook):
    def run(self, response: Response) -> Response:
        def func(self, t: Type) -> Response:
            try:
                self.data = convclass.structure(self.json(), t)
            except TypeError as e:
                fields = dataclasses.fields(t)
                raise TypeError(
                    f"Got miss-matched parameters in dicts. "
                    f"Info about first level:"
                    f"\n\tExpect: {sorted(fields)}"
                    f"\n\tActual: {sorted(dict(self.json()).keys())}"
                    f"\n\nOriginal exception: {e}"
                )
            return self

        response.structure = types.MethodType(func, response)
        return response


ResponseConverterHook = ResponseAttrsConverterHook
